def bubble_sort(arr: list) -> list:
    # sourcery skip: remove-zero-from-range, use-itertools-product
    """Sort a list using bubble sort

    Args:
        arr (list): the list to be sorted

    Returns:
        list: the sorted list
    """
    size = len(arr)

    # Loop over the entire list
    for _ in range(1, size):
        # Loop over the list to swap the items
        # Basically loop in pairs
        for j in range(0, size - 1):
            # Swap items if the current selected is greater than the next
            if arr[j] > arr[j + 1]:
                # Swap
                arr[j], arr[j + 1] = arr[j + 1], arr[j]
    
    return arr


MIN_MERGE = 32

def calc_minrun(size: int):
    """Calculates the minimum size of the runs for a certain number of items.

    Args:
        n (int): the number of items in the array
    """
    r = 0
    while size >= MIN_MERGE:
        r |= size & 1
        size >>= 1
    
    return size + r

def insertion_sort(arr: list, left: int, right: int) -> list:
    """Insertion sort from left index to right index
    Which is at most the size of a run

    Args:
        arr (list): the array to sort
        left (int): the left-most index to start from
        right (int): the right-most index to end from

    Returns:
        list: the sorted list
    """
    # Loop over all items in a run
    for i in range(left + 1, right + 1):
        # Set j equal to the current item
        j = i

        # While we haven't reached the end of the array (on the left)
        # And the item on the right is less than the item on the left
        while j > left and arr[j] < arr[j - 1]:
            # Swap the items and decrease j
            arr[j], arr[j - 1] = arr[j - 1], arr[j]
            j -= 1


def merge(arr: list, left: int, middle: int, right: int):
    """Merge sorted runs

    Args:
        arr (list): the list to be merged
        left (int): the left index of the run
        middle (int): the middle of the run
        right (int): the right index of the run
    """
    # Calculate the length of the left array
    # index in the middle - the left-most index + 1
    left_arr_len = middle - left + 1
    # E.g. with array [0, 1, 2, 3] middle = 2, left = 0
    # 2 - 0 + 1 = 3 from 0 to 2

    # Calculate the length of the right array
    # right-most index - index in the middle
    right_arr_len = right - middle
    # E.g. with array [0, 1, 2, 3] middle = 2, right = 4
    # 4 - 2 = 2 from 3 to 4


    # Create the left and right arrays
    left_arr, right_arr = [], []

    # Add all left items to left array
    left_arr.extend(arr[left + i] for i in range(left_arr_len))
    # Add all items from the right to right array
    right_arr.extend(arr[middle + i + 1] for i in range(right_arr_len))

    # Setup some variables
    l_index, r_index, k = 0, 0, left

    # If there are still items left in both arrays
    # Merge the items
    while l_index < left_arr_len and r_index < right_arr_len:
        # If the left array item is less than the right array item
        if left_arr[l_index] <= right_arr[r_index]:
            # Set the next available item in the original array to the left item
            arr[k] = left_arr[l_index]
            l_index += 1
        else:
            # Set the next available item in the original array to the right item
            arr[k] = right_arr[r_index]
            r_index += 1
    
        # Increase the available array index
        k += 1
    
    # If there are just items left in the left array
    # Add them to the list
    while l_index < left_arr_len:
        arr[k] = left_arr[l_index]
        k += 1
        l_index += 1
    
    # If there are just items left in the right array
    # Add them to the list
    while r_index < right_arr_len:
        arr[k] = right_arr[r_index]
        k += 1
        r_index += 1


def timsort(arr: list) -> list:
    size = len(arr)

    # Calculate the minimum run size
    min_run = calc_minrun(size)

    # Sort each run using insertion sort
    for start in range(0, size, min_run):
        end = min(start + min_run - 1, size - 1)
        insertion_sort(arr, start, end)
    
    # Start merging from size run (or 32)
    # It will merge to form 64, then 128, 256 and so on...
    run_size = min_run
    while run_size < size:
        # Get the starting point of left sub array
        for left in range(0, size, 2 * run_size):
            # Calculate the middle
            # Get the minimum between the last array index and the left index + the run size
            middle = min(size - 1, left + run_size - 1)
            # Calculate the right
            # Get the minimum between the last array index and the left index + 2 * the run size
            # Left index + 2 * run size = the end of the run
            right = min(left + 2 * run_size - 1, size - 1)

            # Merge sub arrays:
            # arr[left..middle] & arr[middle+1..right]
            if middle < right:
                merge(arr, left, middle, right)

        run_size = 2 * run_size
    
    return arr


import sys
sys.path.append("..")

from profq_sorting.helpers.max import get_max


def counting_sort(arr: list, max_value: int = None, place_value: int = None) -> list:
    """Sort a list using counting sort

    Args:
        arr (list): the list to be sorted

    Returns:
        list: the sorted list
    """
    size = len(arr)

    # Get the max value
    if max_value is None:
        max_value = get_max(arr) + 1

    # Set output array and count table
    output = [0] * size
    count = [0] * max_value

    # Count all the numbers
    for i in range(size):
        # Calculate the index for radix sort
        index = arr[i]
        if place_value is not None:
            index = (arr[i] // place_value) % 10
        
        count[index] += 1

    # Cumulative sum
    for i in range(1, max_value):
        count[i] += count[i - 1]

    # Fill the output array
    for i in range(size - 1, -1, -1):
        # Calculate the index for radix sort
        index = arr[i]
        if place_value is not None:
            index = (arr[i] // place_value) % 10

        count[index] -= 1
        # arr[i] is the current element
        # count[arr[i]] is thus the index of the element
        # output[count[arr[i]]] is thus the correct location
        output[count[index]] = arr[i]

    return output


def insertion_sort(arr: list) -> list:
    """Sort a list using insertion sort

    Args:
        arr (list): the list to be sorted

    Returns:
        list: the sorted list
    """
    size = len(arr)

    # Loop over all items
    for i in range(1, size):
        # Get the first unsorted element
        key = arr[i]
        # Index of last sorted element
        j = i - 1

        # While we haven't reached the end of the array
        # And the current sorted item is larger than the unsorted item
        while j >= 0 and arr[j] > key:
            # Move the sorted item to the right
            arr[j + 1] = arr[j]
            # Move back through the array
            j -= 1
        
        # Set the key to the position before the sorted element
        arr[j + 1] = key
    
    return arr


import sys
sys.path.append("..")

from profq_sorting.sorting.insertion_sort import insertion_sort

def bucket_sort(arr: list) -> list:
    # This means there will be 10 buckets
    # From 0 up to 9
    # For 0.0 to 0.999999999...
    slot_num = 10

    # Create an array for each slot
    buckets = [[] for _ in range(slot_num)]

    # Fill the buckets
    for item in arr:
        # Calculate the index of the bucket
        # 0.9 -> 9
        # 0.3452 -> 3
        # Etc
        index = int(slot_num * item)
        # Add the item to the buckets index
        buckets[index].append(item)

    # Sort the buckets
    for i in range(slot_num):
        buckets[i] = insertion_sort(buckets[i])

    output = []

    for i in range(slot_num):
        output.extend(buckets[i])

    return output
    


def shell_sort(arr: list) -> list:
    """Sort a list using shell sort

    Args:
        arr (list): the list to be sorted

    Returns:
        list: the sorted list
    """
    # Calculate the gap to sort the items
    gap = len(arr) // 2

    # While the gap isn't 0
    while gap > 0:
        # Set i to the first element
        i = 0
        # Set j to the element of the gap
        j = gap

        # Loop over all items from gap to arr end
        while j < len(arr):
            # If the ith element is greater, swap the items
            if arr[i] > arr[j]:
                arr[i], arr[j] = arr[j], arr[i]

            # Increase i index and j index to simulatiously
            # Moves both pointers at the same time
            i += 1
            j += 1

            # Set k to the index of the left selected element
            k = i

            # While k is greater than or equal to gap,
            # Make it smaller than gap
            while k >= gap:
                # Check for
                #  left element > right element
                if arr[k - gap] > arr[k]:
                    # Swap if true
                    arr[k - gap], arr[k] = arr[k], arr[k - gap]
                
                # Reduce k to keep it within gap
                k -= 1

        # Make gap smaller
        gap //= 2

    return arr


import sys
sys.path.append("..")

from profq_sorting.sorting.counting_sort import counting_sort
from profq_sorting.helpers.max import get_max

def radix_sort(arr: list) -> list:
    """Sort a list using radix sort

    Args:
        arr (list): the list to be sorted

    Returns:
        list: the sorted list
    """
    # Get the max value
    max_value = get_max(arr)

    # Get the amount of digits (alternate)
    """
    digits = 1
    while max_value > 0:
        max_value /= 10
        digits += 1
    """
    
    place_value = 1
    output = arr

    value = max_value

    # Sort the array
    while value > 0:
        # Use counting sort
        output = counting_sort(output, 10, place_value)

        # Increase place value and decrease digits
        place_value *= 10
        value = max_value // place_value
    
    return output


def selection_sort(arr: list) -> list:
    """Sort a list using selection sort

    Args:
        arr (list): the list to be sorted

    Returns:
        list: the sorted list
    """
    size = len(arr)

    # Loop over entire array except the last element
    for j in range(size - 1):
        # Set the current minimum index to j
        iMin = j

        # Loop over the unsorted array (starting at j + 1)
        for i in range(j + 1, size):
            # Change the minimum when the number is smaller
            if arr[i] < arr[iMin]:
                iMin = i

        # If they aren't the same
        # Swap the minimum and the first element in the unsorted array
        if iMin != j:
            arr[j], arr[iMin] = arr[iMin], arr[j]
    
    return arr